from enum import Enum
from typing import Union

import semver
from pydantic import BaseModel, validator


class BumpTarget(Enum):
    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"


class ReleaseNotes(BaseModel):
    added: list[str] = []
    changed: list[str] = []
    deprecated: list[str] = []
    removed: list[str] = []
    fixed: list[str] = []
    security: list[str] = []

    def __getitem__(self, attr: str) -> list[str]:
        return getattr(self, attr)

    def __setitem__(self, attr: str, value: list[str]) -> None:
        return setattr(self, attr, value)

    def __bool__(self) -> bool:
        return bool(
            self.added
            or self.changed
            or self.deprecated
            or self.removed
            or self.fixed
            or self.security
        )

    @classmethod
    def sections(cls) -> list[str]:
        return list(cls.__fields__.keys())

    def markdown(self) -> str:
        sections = self.dict()
        md = ""
        for name, notes in sections.items():
            if not notes:
                continue

            formatted_notes = "- " + "\n- ".join(notes)
            md += f"""#### {name.title()}

            {formatted_notes}

            """
        return "\n".join(s.lstrip() for s in md.split("\n"))


class VersionInfo(semver.VersionInfo):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @validator("*")
    def validate(cls, v: Union[str, "VersionInfo"]) -> "VersionInfo":
        if isinstance(v, VersionInfo):
            return v
        return cls.parse(v)


class ChangelogUpdate(BaseModel):
    new_version: VersionInfo | None
    old_version: VersionInfo | None
    release_notes: ReleaseNotes



from rich import print
from typer import Option

from changelogger import changelog


def versions(
    latest: bool = Option(
        False,
        "--latest",
        "-l",
        help="Gets the latest version in the changelog file",
    ),
    show_all: bool = Option(
        False,
        "--all",
        "-a",
        help="Get all versions.",
    ),
    start: int = Option(
        0,
        "--start",
        "-s",
        help="What version number to start on.",
    ),
    offset: int = Option(
        10,
        "--offset",
        "-o",
        help="The offset from the start.",
    ),
) -> None:
    """Retrieves the versions found in the changelog file."""
    if latest:
        print(changelog.get_latest_version())
        return

    all_versions = changelog.get_all_versions()
    if show_all:
        print("\n".join(map(str, all_versions)))
        return

    print("\n".join(map(str, all_versions[start : start + offset])))



import typer

from changelogger import changelog
from changelogger.app.prompts import (
    prompt_unreleased_changelog,
    rollback_handler,
)
from changelogger.conf import settings
from changelogger.conf.models import VersionedFile
from changelogger.models.domain_models import ChangelogUpdate

ADDED_HELP = "For new features."
CHANGED_HELP = "For changes in existing functionality."
DEPRECATED_HELP = "For soon-to-be removed features."
REMOVED_HELP = "For now removed features."
FIXED_HELP = "For any bug fixes."
SECURITY_HELP = "In case of vulnerabilities."


def add(
    added: list[str] = typer.Option([], help=ADDED_HELP),
    changed: list[str] = typer.Option([], help=CHANGED_HELP),
    deprecated: list[str] = typer.Option([], help=DEPRECATED_HELP),
    removed: list[str] = typer.Option([], help=REMOVED_HELP),
    fixed: list[str] = typer.Option([], help=FIXED_HELP),
    security: list[str] = typer.Option([], help=SECURITY_HELP),
) -> None:
    """Add changes to the unreleased section of your changelog file. If no
    options are provided, you will be prompted for changes.
    """
    all_versions = changelog.get_all_versions()
    old_version = all_versions[0]
    release_notes = changelog.get_release_notes("Unreleased", old_version)

    update = ChangelogUpdate(
        new_version=None,
        old_version=old_version,
        release_notes=release_notes,
    )

    if added or changed or deprecated or removed or fixed or security:
        # Update the existing unreleased changes with the added notes
        update.release_notes.added.extend(added)
        update.release_notes.changed.extend(changed)
        update.release_notes.deprecated.extend(deprecated)
        update.release_notes.removed.extend(removed)
        update.release_notes.fixed.extend(fixed)
        update.release_notes.security.extend(security)
    else:
        update = prompt_unreleased_changelog(update)

    changelog_file = VersionedFile(
        rel_path=settings.CHANGELOG_PATH,
        pattern=settings.OVERVIEW_JINJA_PATTERN,
        jinja_rel_path=settings.OVERVIEW_JINJA_PATH,
    )

    with rollback_handler():
        changelog.update_versioned_files(
            update=update,
            versioned_files=[changelog_file],
        )



from functools import partial, wraps
from pathlib import Path
from typing import Callable

import typer
from git.exc import InvalidGitRepositoryError
from git.repo import Repo
from rich import print

from changelogger.app.commands.add import add
from changelogger.app.commands.check import check
from changelogger.app.commands.init import init
from changelogger.app.commands.notes import notes
from changelogger.app.commands.upgrade import upgrade
from changelogger.app.commands.versions import versions
from changelogger.conf import settings


class App(typer.Typer):
    def add_command(
        self,
        cmd: Callable,
        alias: str | None = None,
    ) -> None:
        """Adds the provided callable as command on this App instance.
        If an alias is provided, the command can be invoked with that
        name as well.
        """
        self.command()(cmd)

        if not alias:
            return

        alias_cmd = wraps(cmd)(partial(cmd))
        alias_cmd.__doc__ = (
            f"Alias for the `{cmd.__name__}` command. {cmd.__doc__}"
        )

        # Alias commands
        self.command(alias, rich_help_panel="Aliases")(alias_cmd)


app = App()

app.add_command(add)
app.add_command(check, "ch")
app.add_command(init)
app.add_command(notes)
app.add_command(upgrade, "up")
app.add_command(versions)


def version_callback(value: bool):
    if not value:
        return

    print(settings.CHANGELOGGER_VERSION)
    raise typer.Exit()


@app.callback()
def changelogger(
    ctx: typer.Context,
    _: bool = typer.Option(
        False,
        "-v",
        "--version",
        help="The version of Changelogger you have installed.",
        callback=version_callback,
    ),
):
    if (
        not ctx.invoked_subcommand == init.__name__
        and not settings.CHANGELOG_PATH.exists()
    ):
        print(
            "[bold red]Error: [/bold red]"
            f"Could not find changelog file [bold]{settings.CHANGELOG_PATH}[/bold]."
        )
        raise typer.Abort()

    if settings.HAS_DEFAULTS:
        try:
            __ = Repo(Path.cwd()).git_dir
        except InvalidGitRepositoryError:
            print(
                "[bold red]Error: [/bold red]"
                "Must be in a git repo to use the default behavior."
            )
            raise typer.Abort()


changelogger.__doc__ = f"""
Automated management of your changelog and other versioned files, following the
principles of Keep a Changelog and Semantic Versioning.\n
version: {settings.CHANGELOGGER_VERSION}
"""
changelogger.__doc__ += "\nIN DEBUG MODE" if settings.DEBUG else ""



from pathlib import Path

from git.repo import Repo


class _Git:
    KEY_URL = "url"
    GITHUB_HTTPS_PREFIX = "https://github.com/"
    GITHUB_SSH_PREFIX = "git@github.com:"
    GIT_SUFFIX = ".git"

    def __init__(self) -> None:
        self._repo = Repo(
            Path.cwd(),
        )

    def _get_git_repo(self) -> str:
        git_url = self._repo.remotes.origin.config_reader.get(
            self.KEY_URL,
        )
        return (
            git_url.lstrip(
                self.GITHUB_HTTPS_PREFIX,
            )
            .lstrip(
                self.GITHUB_SSH_PREFIX,
            )
            .rstrip(
                self.GIT_SUFFIX,
            )
        )

    def _get_first_commit(self) -> str:
        commits = list(self._repo.iter_commits())
        return commits[-1].hexsha[:10]

    def get_ctx(self) -> dict:
        return dict(
            repo=self._get_git_repo(),
            first_commit=self._get_first_commit(),
        )


get_ctx = _Git().get_ctx



from importlib import resources

from changelogger.conf.defaults import *
from changelogger.conf.models import ChangeloggerConfig

_config = ChangeloggerConfig.from_config_or_default()

DEBUG = False

CHANGELOG_PATH = _config.changelog.rel_path

CHANGELOG_JINJA = resources.files("changelogger").joinpath(
    "assets/changelog.jinja2",
)

OVERVIEW_JINJA_PATTERN = _config.changelog.overview.pattern
OVERVIEW_JINJA_PATH = _config.changelog.overview.jinja_rel_path

LINKS_JINJA_PATTERN = _config.changelog.links.pattern
LINKS_JINJA_PATH = _config.changelog.links.jinja_rel_path

VERSIONED_FILES = _config.versioned_files
VERSIONED_FILES.extend(_config.changelog.as_versioned_files())

HAS_DEFAULTS = _config.changelog.has_defaults()

# THIS MUST BE IMPORTED LAST
try:
    from changelogger.conf._settings_override import *  # nopycln: import
except:
    pass



import typer
from rich import print
from rich.markdown import Markdown

from changelogger import changelog
from changelogger.app.prompts import (
    prompt_unreleased_changelog,
    rollback_handler,
)
from changelogger.conf import settings
from changelogger.models.domain_models import BumpTarget, ChangelogUpdate


def upgrade(
    version_to_bump: BumpTarget,
    confirm: bool = typer.Option(
        True,
        help="Confirm the release notes before applying them.",
    ),
    prompt_changelog: bool = typer.Option(
        True,
        help="Prompt for additional release notes before applying them.",
    ),
) -> None:
    """Upgrades all versioned files, as specified in the changelogger config file."""
    old_version = changelog.get_latest_version()
    bump = getattr(old_version, f"bump_{version_to_bump.value}")
    new_version = bump()

    release_notes = changelog.get_release_notes("Unreleased", old_version)
    update = ChangelogUpdate(
        old_version=old_version,
        new_version=new_version,
        release_notes=release_notes,
    )

    if prompt_changelog:
        update = prompt_unreleased_changelog(update)

    print(f"Upgrading {old_version} ==> {new_version}")
    md = f"\n# Changelog updates for [{new_version}]\n"
    if update_md := update.release_notes.markdown():
        md += update_md
    else:
        md += "\n*No notes found or added*"

    md += "\n---\n"
    print(Markdown(md))

    if confirm:
        typer.confirm("Do these changes look correct?", abort=True)

    with rollback_handler():
        changelog.update_versioned_files(
            update,
            settings.VERSIONED_FILES,
        )



from collections import defaultdict
from pathlib import Path
from typing import DefaultDict

import yaml  # type: ignore
from pydantic import BaseModel

from changelogger.conf import git
from changelogger.conf.defaults import (
    CHANGELOGGER_PATH,
    DEFAULT_CHANGELOG_PATH,
    DEFAULT_LINKS_JINJA_PATH,
    DEFAULT_LINKS_JINJA_PATTERN,
    DEFAULT_OVERVIEW_JINJA_PATH,
    DEFAULT_OVERVIEW_JINJA_PATTERN,
)


class VersionedFile(BaseModel):
    rel_path: Path
    pattern: str
    jinja: str | None = None
    jinja_rel_path: Path | None = None
    context: dict = {}

    def simple_dict(self) -> dict:
        return {
            k: v if not isinstance(v, Path) else str(v)
            for k, v in self.dict().items()
            if v
        }


class ChangelogSegment(BaseModel):
    pattern: str
    jinja_rel_path: Path


_default_overview = ChangelogSegment(
    pattern=DEFAULT_OVERVIEW_JINJA_PATTERN,
    jinja_rel_path=DEFAULT_OVERVIEW_JINJA_PATH,
)

_default_links = ChangelogSegment(
    pattern=DEFAULT_LINKS_JINJA_PATTERN,
    jinja_rel_path=DEFAULT_LINKS_JINJA_PATH,
)


class Changelog(BaseModel):
    rel_path: Path = Path(DEFAULT_CHANGELOG_PATH)
    overview: ChangelogSegment = _default_overview
    links: ChangelogSegment = _default_links

    def has_defaults(self) -> bool:
        return (
            self.overview == _default_overview or self.links == _default_links
        )

    def as_versioned_files(self) -> list[VersionedFile]:
        context: DefaultDict = defaultdict(dict)
        if self.has_defaults():
            context["git"] = git.get_ctx()

        return [
            VersionedFile(
                rel_path=self.rel_path,
                pattern=segment.pattern,
                jinja_rel_path=segment.jinja_rel_path,
                context=context,
            )
            for segment in (self.overview, self.links)
        ]


class ChangeloggerConfig(BaseModel):
    changelog: Changelog = Changelog()
    versioned_files: list[VersionedFile] = []

    @classmethod
    def from_config_or_default(cls) -> "ChangeloggerConfig":
        if not CHANGELOGGER_PATH.exists():
            return cls()

        raw_config = yaml.safe_load(
            CHANGELOGGER_PATH.read_text(),
        )
        return cls(**raw_config)



import re
from datetime import date
from pathlib import Path
from textwrap import dedent
from typing import Any

import typer
import yaml  # type: ignore
from rich import print
from rich.console import Console
from rich.panel import Panel
from rich.syntax import Syntax

from changelogger import changelog
from changelogger.conf import git, settings
from changelogger.conf.models import VersionedFile
from changelogger.templating import render_jinja
from changelogger.utils import cached_compile


def init(
    prompt_changelog: bool = True,
    prompt_versioned_files: bool = True,
):
    """Setup a git repository to work with Changelogger or reinitialize
    an existing one.
    """

    console = Console()
    if prompt_changelog:
        _init_changelog(console)

    if prompt_versioned_files:
        _init_changelogger(console)


def _init_changelog(console: Console) -> None:
    # If they don't already have a changelog file or want to create a new one
    changelog_exists = settings.CHANGELOG_PATH.exists()
    if changelog_exists:
        if not typer.confirm(
            "It looks like you've already specified your changelog "
            f'as "{settings.CHANGELOG_PATH}"; are you sure you want to '
            "create a new one?"
        ):
            return
    elif not typer.confirm(
        f'Would you like to generate a "{settings.CHANGELOG_PATH}" file?'
    ):
        return

    # Wants to replace/create a CHANGELOG.md
    settings.DEFAULT_CHANGELOG_PATH.write_text(
        render_jinja(
            settings.CHANGELOG_JINJA.read_text(),
            dict(
                today=date.today(),
                context=dict(git=git.get_ctx()),
            ),
        )
    )
    console.clear()

    console.print(f'"{settings.CHANGELOG_PATH}" successfully created!')


def _init_changelogger(console: Console) -> None:
    # If they don't already have a changelogger file or want to create a new one
    changelogger_exists = settings.CHANGELOGGER_PATH.exists()
    if changelogger_exists:
        if not typer.confirm(
            "It looks like you've already specified your changelogger "
            f'as "{settings.CHANGELOGGER_PATH}"; are you sure you want to '
            "replace it?"
        ):
            return

    if not typer.confirm(
        "Would you like Changelogger to monitor and update any files in "
        'addition to "CHANGELOG.md"?'
    ):
        return

    versioned_files: list[VersionedFile] = []
    while True:
        if versioned_file := _prompt_versioned_file(console):
            versioned_files.append(versioned_file)

        if not typer.confirm("Any other versioned files?"):
            break

    settings.CHANGELOGGER_PATH.write_text(
        yaml.safe_dump(
            dict(
                versioned_files=[
                    versioned_file.simple_dict()
                    for versioned_file in versioned_files
                ],
            ),
        )
    )
    console.clear()

    console.print(f'"{settings.CHANGELOGGER_PATH}" successfully created!')


def _prompt_versioned_file(console: Console) -> VersionedFile | None:
    console.clear()
    rel_path = typer.prompt(
        "What is the relative path of this versioned file?", type=Path
    )
    if not rel_path.exists():
        print(f'[bold]Could not find[/bold] "{rel_path}"; Skipping this file.')
        return None

    console.clear()
    pattern = typer.prompt(
        "What pattern should changelogger use to find and replace the "
        "versioned information?"
    )

    # TODO
    # This doesn't encapsulate state which will be available during the actual
    # Upgrade process. This confirmation could be more accurate.
    rendered_pattern = render_jinja(
        pattern,
        dict(
            old_version=changelog.get_latest_version(),
            today=date.today(),
        ),
    )
    num_matches = len(re.findall(rendered_pattern, rel_path.read_text()))
    match = cached_compile(rendered_pattern).search(rel_path.read_text())
    if not match:
        print(
            f'Could not find a match in "{rel_path}" using the rendered '
            f"pattern [bold]{rendered_pattern}[/bold]; skipping this file."
        )
        return None

    console.clear()
    num_matches = len(re.findall(rendered_pattern, rel_path.read_text()))
    title = f"{num_matches} Match{'es' if num_matches > 1 else ''} Found!"

    print(
        Panel.fit(
            match[0],
            title=f"[bold green]{title}[/bold green]",
            subtitle=":backhand_index_pointing_up: First Match",
            padding=1,
        )
    )

    context = _prompt_context()
    console.clear()

    metadata = yaml.safe_dump(
        dict(
            rel_path=str(rel_path),
            pattern=pattern,
            context=context or None,
        )
    )
    print(
        Panel.fit(
            Syntax(metadata, "yaml"),
            title=f"[bold]{rel_path} Metadata[/bold]",
            padding=1,
        )
    )

    if not typer.confirm(
        "Does everything look good? If so, we'll prompt you for the "
        "replacement jinja."
    ):
        print("Skipping this file.")
        return None

    console.clear()
    jinja = _prompt_jinja(rel_path, pattern, context)
    if not jinja:
        print("No jinja entered; skipping this file.")
        return None

    console.clear()
    jinja_rel_path = None
    if "\n" in jinja:
        if typer.confirm(
            "It looks like your jinja is multiple lines; "
            "do you want to save this in its own file?"
        ):
            default_path = Path(f"./.changelogger/{rel_path}.jinja2")
            jinja_rel_path = typer.prompt(
                "Where do you want to save the template?",
                default=default_path,
                type=Path,
            )

            jinja_rel_path.parent.mkdir(parents=True, exist_ok=True)
            jinja_rel_path.write_text(jinja)
            print(f'Successfully saved jinja to "{jinja_rel_path}"!')

    return VersionedFile(
        rel_path=rel_path,
        pattern=pattern,
        jinja=jinja if not jinja_rel_path else None,
        jinja_rel_path=jinja_rel_path,
        context=context,
    )


def _prompt_context() -> dict[str, Any]:
    if not typer.confirm(
        "Do you want to add context to use in your jinja " "template?"
    ):
        return {}

    CONTEXT_MARKER = dedent(
        """
        # Provide any context in the above context object to make it available
        # in your jinja file.
        #
        # -- EXAMPLE --
        # context:
        #   pet_names:
        #   - friday
        #   - lucy
        #
        # -- USAGE --
        # {% for pet_name in  context.pet_names %}
        # ...
        # {% endfor %}
        """
    )
    ctx_str = typer.edit("context:\n\n" + CONTEXT_MARKER, extension=".yml")
    try:
        return yaml.safe_load(ctx_str) if ctx_str else {}
    except yaml.error.YAMLError:
        print("Unable to extract context; Skipping context.")
        return {}


def _prompt_jinja(
    rel_path: Path,
    pattern: str,
    context: dict[str, Any],
) -> str | None:
    MARKER = (
        dedent(
            f"""
        {{#
        Enter the jinja replacement above.

        rel_path: {rel_path}
        pattern: {pattern}
        """
        )
        + yaml.safe_dump(dict(context=context or None))
        + "#}"
    )

    message = typer.edit("\n\n" + MARKER, extension=".jinja2")
    if not message:
        print("No jinja supplied. Skipping file.")
        return None

    return message.split(MARKER)[0].rstrip()



from pathlib import Path
from typing import Literal

from changelogger import templating
from changelogger.conf import settings
from changelogger.conf.models import VersionedFile
from changelogger.exceptions import (
    CommandException,
    RollbackException,
    UpgradeException,
)
from changelogger.models.domain_models import (
    ChangelogUpdate,
    ReleaseNotes,
    VersionInfo,
)
from changelogger.utils import cached_compile

CHANGELOG_PARTITION_RELEASE_NOTES = "RELEASE NOTES"
CHANGELOG_PARTITION_LINKS = "LINKS"


def _get_changelog_parition(partition: str) -> str:
    changelog_content = settings.CHANGELOG_PATH.read_text()

    start_partition = f"<!-- BEGIN {partition} -->"
    end_partition = f"<!-- END {partition} -->"
    partition_re = cached_compile(
        rf"{start_partition}([\s\S]*){end_partition}"
    )

    match = partition_re.search(changelog_content)
    if not match:
        raise CommandException(
            f"Expected partition for `{partition}`; None found."
        )
    return match[1]


def _get_release_notes_parition() -> str:
    return _get_changelog_parition(CHANGELOG_PARTITION_RELEASE_NOTES)


def _get_links_parition() -> str:
    return _get_changelog_parition(CHANGELOG_PARTITION_LINKS)


def get_all_links() -> dict[VersionInfo | str, str]:
    lines = _get_links_parition().split("\n")

    links = {}
    for line in lines:
        match = cached_compile(
            r"\[(.*)]: (.*)",
        ).search(
            line,
        )

        if not match:
            continue

        version_str = match[1]
        link = match[2]

        if version_str == "Unreleased":
            links[version_str] = link
            continue

        match = VersionInfo._REGEX.fullmatch(version_str)
        if not match:
            continue

        links[VersionInfo.parse(version_str)] = link

    return links


def get_all_versions() -> list[VersionInfo]:
    lines = _get_release_notes_parition().split("\n")
    versions = []
    for line in lines:
        match = cached_compile(
            r"### \[(.*)]",
        ).search(
            line,
        )

        if not match:
            continue

        version_str = match[1]

        match = VersionInfo._REGEX.fullmatch(version_str)
        if not match:
            continue

        versions.append(VersionInfo.parse(version_str))
    return versions


def get_sorted_versions() -> list[VersionInfo]:
    return sorted(get_all_versions())


def get_latest_version() -> VersionInfo:
    versions = get_sorted_versions()
    if not versions:
        raise UpgradeException(f"This changelog has no versions currently.")
    return versions[-1]


def get_release_notes(
    new_version: VersionInfo | Literal["Unreleased"],
    old_version: VersionInfo | None,
) -> ReleaseNotes:
    new_version_pattern = str(new_version).replace(".", r"\.")

    content = _get_release_notes_parition()

    pattern = rf"### \[{new_version_pattern}\]( - \d+-\d+-\d+)?([\s\S]*)"
    if old_version:
        old_version_pattern = str(old_version).replace(".", r"\.")
        pattern += rf"### \[{old_version_pattern}\]"
    match = cached_compile(
        pattern,
    ).search(
        content,
    )
    if not match:
        raise UpgradeException("Could not extract release notes.")

    raw_notes = match[2]
    raw_sections = cached_compile("[#]+").split(raw_notes)

    release_notes = ReleaseNotes()
    for section in raw_sections:
        section_name, *notes = section.split("\n")
        if not section_name:
            continue

        notes = [
            match[1].strip()
            for note in notes
            if (
                match := cached_compile(
                    r"\- (.*)",
                ).search(note)
            )
        ]

        attr = section_name.strip().lower()
        release_notes[attr] = notes

    return release_notes


def _rollback(rollback: list[tuple[Path, str]]) -> None:
    for path, content in rollback:
        path.write_text(content)


def update_versioned_files(
    update: ChangelogUpdate,
    versioned_files: list[VersionedFile],
) -> None:
    rollback: list[tuple[Path, str]] = []
    try:
        for file in versioned_files:
            content = file.rel_path.read_text()
            rollback.append((file.rel_path, content))
            new_content = templating.update(file, update, content)
            file.rel_path.write_text(new_content)
    except Exception as upgrade_exc:
        try:
            # Need to reverse rollback list for proper rollback
            _rollback(rollback[::-1])
        except Exception as rollback_exc:
            raise RollbackException(
                "An exception occured while upgrading; rollback unsuccessful."
            ) from rollback_exc

        raise UpgradeException(
            "An exception occured while upgrading; rollback successful."
        ) from upgrade_exc



from contextlib import contextmanager

from rich import print
from rich.markdown import Markdown

from changelogger.exceptions import RollbackException, UpgradeException
from changelogger.models.domain_models import ChangelogUpdate


def prompt_unreleased_changelog(update: ChangelogUpdate) -> ChangelogUpdate:
    for name, notes in update.release_notes.dict().items():
        done = False
        print(Markdown(f"## Updating **{name.title()}**"))
        while not done:
            print(f"Any further additions for [bold]{name.title()}[/bold]?")
            print(notes)
            if new_note := input("New note [Enter to continue]: "):
                notes.append(new_note)
            else:
                done = True
        update.release_notes[name] = notes
    return update


@contextmanager
def rollback_handler():
    try:
        yield
    except UpgradeException as e:
        print(f"[bold red]Failed to update.[/bold red] {str(e)}")
        if isinstance(e, RollbackException):
            note = "MANUAL INTERVENTION REQUIRED to fix versioned files."
            print(f"\n[bold red]{note}[/bold red]")



import typer
from rich import print
from rich.markdown import Markdown

from changelogger import changelog
from changelogger.models.domain_models import ReleaseNotes, VersionInfo

VERSION_HELP = """
The version notes you would like to retrieve.

\b
\nAdditionally, you use these shortcuts.
\nlatest: Retrieves the latest versions notes.
\nunreleased: Retrieves any unreleased notes. This is the default behaviour
if no value is provided.

"""


def notes(
    version: str = typer.Argument(
        "unreleased",
        help=VERSION_HELP,
    ),
    pretty: bool = typer.Option(
        True,
        help="Prints the content in a more legible format if selected.",
    ),
) -> None:
    """Retrieves the changelog content for the specified version."""
    prefix = "# " if pretty else ""

    release_notes = None
    if version == "unreleased":
        release_notes = (
            _unreleased_notes().markdown()
            or f"{prefix}There are no unreleased changes."
        )

    if release_notes is None:
        parsed_version = (
            changelog.get_latest_version()
            if version == "latest"
            else VersionInfo.parse(version)
        )
        release_notes = (
            _released_notes(parsed_version).markdown()
            or f"{prefix}There are no notes for version {version}."
        )

    print(Markdown(release_notes) if pretty else release_notes)


def _unreleased_notes() -> ReleaseNotes:
    all_versions = changelog.get_all_versions()

    # all_versions[0] is the topmost version in the changelog file.
    return changelog.get_release_notes("Unreleased", all_versions[0])


def _released_notes(version: VersionInfo) -> ReleaseNotes:
    all_versions = changelog.get_all_versions()
    if version not in all_versions:
        print(f"Could not find version {version}.")
        raise typer.Abort()

    i = all_versions.index(version)
    (prev_version,) = all_versions[i + 1 : i + 2] or (None,)
    return changelog.get_release_notes(version, prev_version)



from importlib import resources
from pathlib import Path

DEFAULT_CHANGELOG_PATH = Path("CHANGELOG.md")

DEFAULT_OVERVIEW_JINJA_PATTERN = (
    r"### \[Unreleased\]([\s\S]*)### \[{{ old_version }}]"
)
DEFAULT_LINKS_JINJA_PATTERN = r"\[Unreleased\]:.*\n"

DEFAULT_OVERVIEW_JINJA_PATH = resources.files("changelogger").joinpath(
    "assets/.cl.overview.jinja2"
)
DEFAULT_LINKS_JINJA_PATH = resources.files("changelogger").joinpath(
    "assets/.cl.links.jinja2"
)

CHANGELOGGER_NAME = ".changelogger.yml"

CHANGELOGGER_PATH = (
    (p1 := Path(CHANGELOGGER_NAME)).exists()
    and p1
    or (p2 := Path(".changelogger/").joinpath(CHANGELOGGER_NAME)).exists()
    and p2
    or (p3 := Path(".github/").joinpath(CHANGELOGGER_NAME)).exists()
    and p3
    or p1
)

CHANGELOGGER_VERSION = "0.11.3"



from collections import Counter
from typing import Callable

import typer
from rich import print
from rich.progress import Progress

from changelogger import changelog, templating
from changelogger.conf import settings
from changelogger.conf.models import VersionedFile
from changelogger.exceptions import ValidationException
from changelogger.models.domain_models import ChangelogUpdate
from changelogger.utils import cached_compile


def check(
    sys_exit: bool = typer.Option(
        False,
        "--sys-exit",
        "--fail",
        help="Exit with a status of 2 if any versioned files are invalid.",
    ),
    files: list[str] = typer.Option(
        [],
        "--file",
        help="Only check the specified file(s).",
    ),
) -> None:
    """Checks the versioned files for any unparsable sections which do not match
    the Changelogger configuration and reports them.
    """
    versioned_files = (
        settings.VERSIONED_FILES
        if not (files_set := set(files))
        else [
            file
            for file in settings.VERSIONED_FILES
            if str(file.rel_path) in files_set
        ]
    )
    try:
        _check_versioned_files(versioned_files)
    except ValidationException as e:
        print(f"[bold red]Error:[/bold red] {str(e)}")
        if sys_exit:
            raise typer.Exit(code=2)
    else:
        print(
            ":white_heavy_check_mark: [bold green]Versioned files are valid![/bold green]"
        )


def _check_versioned_files(versioned_files: list[VersionedFile]) -> None:
    """For each of the user-specified and default versioned files, check
    that a search for the pattern over the files content results in a find.
    """

    # Contrive a fake update so we can check if the pattern would have been
    # found.
    old_version = changelog.get_latest_version()
    update = ChangelogUpdate(
        new_version=old_version.bump_minor(),
        old_version=old_version,
        release_notes=changelog.get_release_notes("Unreleased", old_version),
    )

    counts = Counter(file.rel_path for file in versioned_files)
    with Progress() as progress:
        if settings.CHANGELOG_PATH in counts:
            counts[settings.CHANGELOG_PATH] += 6

        tasks = {
            path: progress.add_task(
                f'Checking [green]"{path}"[/green]',
                total=total,
            )
            for path, total in counts.items()
        }

        for file in versioned_files:
            _check_versioned_file(file, update)
            progress.advance(tasks[file.rel_path])

        if any(
            file.rel_path == settings.CHANGELOG_PATH
            for file in versioned_files
        ):
            advancer = lambda: progress.advance(
                tasks[settings.CHANGELOG_PATH],
            )
            _check_changelog(advancer)


def _check_versioned_file(
    file: VersionedFile, update: ChangelogUpdate
) -> None:
    """Renders the versioned files pattern with an update and confirms
    there's a match.
    """

    pattern = templating.render_pattern(file, update)
    content = file.rel_path.read_text()
    if cached_compile(pattern).search(content):
        return

    raise ValidationException(
        f"Could not find the pattern `[bright_blue]{file.pattern}[/bright_blue]` "
        f'in "{file.rel_path}".\n\n'
        f"Rendered pattern used when searching: `[bright_blue]{pattern}[/bright_blue]`."
    )


def _check_changelog(advance: Callable) -> None:
    """Validates the specified versioned files are parsable and updatable."""
    # Validate there's at least 1 version
    # Point of Failure 0
    all_versions = changelog.get_all_versions()
    if not all_versions:
        raise ValidationException(
            "Expected there to be at least 1 version; None found."
        )
    advance()

    # Validate all release notes are parseable for all versions
    # Point of Failure 1
    changelog_versions = ["Unreleased", *all_versions, None]
    for new_version, old_version in zip(
        changelog_versions, changelog_versions[1:]
    ):
        try:
            changelog.get_release_notes(new_version, old_version)  # type: ignore
        except Exception as e:
            raise ValidationException(
                f"Failed to validate notes for version {new_version}: {str(e)}."
            )
    advance()

    # Validate there are links in the expected format for all versions
    # Point of Failure 2
    sorted_versions = changelog.get_sorted_versions()
    all_links = changelog.get_all_links()
    for prev_version, version in zip(sorted_versions, sorted_versions[1:]):
        link = all_links.get(version)
        if not link:
            raise ValidationException(
                f"Could not find the link for version {version}"
            )

        # Point of Failure 3
        if f"{prev_version}...{version}" not in link:
            raise ValidationException(
                f"Link is incorrect for version {version}"
            )
    advance()

    link = all_links.get("Unreleased")
    # Point of Failure 4
    if not link:
        raise ValidationException(
            "Could not find the link for unreleased changes."
        )
    advance()

    # Point of Failure 5
    if f"{all_versions[0]}...HEAD" not in link:
        raise ValidationException(
            "Link is incorrect for the unreleased changes."
        )
    advance()

    # Point of Failure 6
    if sorted_versions[0] not in all_links:
        raise ValidationException(
            f"Could not find the link for version {sorted_versions[0]}"
        )
    advance()



from datetime import date
from functools import partial
from re import Match
from typing import Any

from jinja2 import BaseLoader, Environment, Template

from changelogger.conf.models import VersionedFile
from changelogger.models.domain_models import ChangelogUpdate
from changelogger.utils import cached_compile


def update(
    file: VersionedFile,
    update: ChangelogUpdate,
    content: str,
) -> str:
    """Replaces the versioned files rendered pattern in the supplied content."""

    replacement_str = file.jinja
    if not replacement_str:
        assert file.jinja_rel_path, "No valid jinja template found."
        replacement_str = file.jinja_rel_path.read_text()

    var_getter = partial(_get_variables, file, update)
    pattern = render_jinja(file.pattern, var_getter())

    # re.sub can take a callable as the replacement argument rather than a
    # string. This callable accepts a match and returns a string. For each
    # match which is found, re.sub will call repl with the match, and
    # replace the found pattern with the output string of the user supplied
    # repl function.
    repl = lambda m: render_jinja(replacement_str, var_getter(m))
    return cached_compile(pattern).sub(repl, content)


def render_pattern(
    file: VersionedFile,
    update: ChangelogUpdate,
) -> str:
    variables = _get_variables(file, update)
    return render_jinja(file.pattern, variables)


def render_jinja(tmpl: str, variables: dict[str, Any]) -> str:
    return _tmpl(tmpl).render(**variables)


def _tmpl(jinja: str) -> Template:
    template_env = Environment(loader=BaseLoader())
    return template_env.from_string(jinja)


def _get_variables(
    versioned_file: VersionedFile,
    update: ChangelogUpdate,
    match: Match | None = None,
) -> dict[str, Any]:
    return dict(
        new_version=update.new_version,
        old_version=update.old_version,
        today=date.today(),
        sections=update.release_notes.dict(),
        context=versioned_file.context,
        match=match,
    )
